/*
 * Copyright (c) 2002, ERIK WISTRAND
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer. 
 * 
 * - Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution. 
 * 
 * - Neither the name of the VALENTINE MIDLET PROJECT nor the names of its 
 *   contributors may be used to endorse or promote products derived 
 *   from this software without specific prior written permission. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package com.wistrand.midlet.sprite;

import javax.microedition.lcdui.*;
import javax.microedition.midlet.*;

import java.util.*;

import com.wistrand.midlet.sprite.*;
import com.wistrand.midlet.fp.*;


/**
 * Canvas which paints sprites.
 */
public class SpriteCanvas extends Canvas implements Runnable, SpriteHandler  {

  public Display  display;
  public Vector   sprites = new Vector();

  public int      bgColor = 0xffffff;

  boolean  bRun     = false;
  Thread   runner   = null;
  boolean  bPaused  = false;

  Image    offImage    = null;
  Graphics offGraphics = null;

  public Image    bgImage     = null;
  public Graphics bgGraphics  = null;
  public boolean  bUseBgImage = false;

  public int      nShowInfo   = 0;
  public static int MAXINFO   = 2;

  public KeyListener kl;

  public SpriteCanvas(Display display, KeyListener kl) {
    this.display = display;
    this.kl      = kl;

    offImage    = Image.createImage(getWidth(), getHeight());
    offGraphics = offImage.getGraphics();

  }
  


  int bgX = 0;
  int bgY = 0;
  int bgStep = 10;

  void makeBgImage() {
    if(bgImage == null) {
      bgY = bgX  = 0;
      bgImage    = Image.createImage(getWidth(), getHeight());
      bgGraphics = bgImage.getGraphics();
      paintBg(bgGraphics);

      bgStep = bgImage.getWidth() / 16;
    }

    int r0 = (bgColor >> 16) & 0xff;
    int g0 = (bgColor >> 8) & 0xff;
    int b0 = (bgColor) & 0xff;

    if(bgY < bgImage.getHeight()) {
      int R = Math.min(255, 0 + 255 * bgX / bgImage.getWidth());
      int G = Math.min(255, 0 + 255 * bgY / bgImage.getHeight());
      int B = Math.min(255, b0 + bgColor & 0xff);
      
      int pix = (R << 16) | (G << 8) | B;
      bgGraphics.setColor(pix);
      bgGraphics.fillRect(bgX, bgY, bgStep, bgStep);
      
      bgX += bgStep;
      if(bgX >= bgImage.getWidth()) {
	bgY += bgStep;
	bgX = 0;
      }
    }
  }


  public void keyPressed(int keyCode) {
    int gameAction = getGameAction(keyCode);

    if(kl != null) {
      kl.keyPressed(keyCode, gameAction);
    }
  }

  public void clearSprites() {
    sprites.setSize(0);
  }

  public void setBgColor(int col) {
    bgColor = col;
  }

  public void togglePause() {
    bPaused = !bPaused;
  }

  public boolean isRunning() {
    return runner != null && bRun;
  }

  boolean isPainting = false;

  public void paint(Graphics g) {
    isPainting = true;
    g.drawImage(offImage, 0, 0, Graphics.TOP | Graphics.LEFT);
    isPainting = false;
  }
  
  private void paintAll(Graphics g) {
    
    if(bUseBgImage) {
      makeBgImage();
      g.drawImage(bgImage, 0, 0, Graphics.TOP | Graphics.LEFT);
    } else {
      paintBg(g);
    }
    paintSprites(g);
    if(nShowInfo > 0) {
      paintInfo(g);
    }
  }
  

  void paintSprites(Graphics g) {
    for(int i = 0; i < sprites.size(); i++) {
      Sprite s  = (Sprite)sprites.elementAt(i);
      s.paint(g);
    }
  }
  
  void updateSprites() {
    for(int i = 0; i < sprites.size(); i++) {
      Sprite s  = (Sprite)sprites.elementAt(i);
      s.update();
    }
  }
  
  
  void paintBg(Graphics g) {
    g.setColor(bgColor);
    g.fillRect(0, 0, getWidth(), getHeight());
  }

  void paintInfo(Graphics g) {
    g.setColor(bgColor ^ 0xffffff);

    int dy = 8;
    int y = 5;
    int x = 5;

    g.drawString("ms/frame: " + updateTime,x,y,Graphics.TOP|Graphics.LEFT);
    y += dy;
    g.drawString("delay: "  + delay,    x,y,Graphics.TOP|Graphics.LEFT); 
    y += dy;
    g.drawString("sprites: " +sprites.size(),x,y,Graphics.TOP|Graphics.LEFT);
    y += dy;

    if(nShowInfo > 1) {
      Runtime rt = Runtime.getRuntime();
      g.drawString("free: "   + rt.freeMemory(),x,y,Graphics.TOP|Graphics.LEFT);
      y += dy;
      g.drawString("total: "  + rt.totalMemory(), x, y, Graphics.TOP | Graphics.LEFT);
      y += dy;
    }
  }

  public void add(Sprite s) {
    sprites.addElement(s);
  }

  public void addAfter(Sprite s, Sprite before) {
    int ix = sprites.indexOf(before);
    if(ix != -1) {
      sprites.insertElementAt(s, ix + 1);
    } else {
      add(s);
    }
  }

  public void addBefore(Sprite s, Sprite before) {
    int ix = sprites.indexOf(before);
    if(ix > 0) {
      sprites.insertElementAt(s, ix - 1);
    } else {
      sprites.insertElementAt(s, 0);
    }
  }

  public void remove(Sprite s) {
    sprites.removeElement(s);
  }

  

  long targetTime  = 1000 / 30;
  long delay       = 1;
  long updateTime  = 0;
  long minDelay    = 10;

  public void run() {
    while (bRun) {
      try {
	if(bPaused) {
	  Thread.sleep(500);
	} else {
	  long startTime = System.currentTimeMillis();

	  Thread.sleep(delay);

	  // update sprites and paint to off buffer
	  updateSprites();
	  paintAll(offGraphics);

	  // ...and trigger a repaint event
	  repaint();
	  updateTime = System.currentTimeMillis() - startTime;

	  long delta = targetTime - updateTime;
	  delay = Math.max(minDelay, delta);
	}
      } catch (Exception e) { 
	e.printStackTrace();
      }
    }
  }

 
  public void stop() {
    bRun = false;
  }

  
  public void start() {
    if(runner == null) {
      runner = new Thread(this);
      bRun = true;
      runner.start();
    }
  }
}

