/*
 * Copyright (c) 2002, ERIK WISTRAND
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer. 
 * 
 * - Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution. 
 * 
 * - Neither the name of the VALENTINE MIDLET PROJECT nor the names of its 
 *   contributors may be used to endorse or promote products derived 
 *   from this software without specific prior written permission. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package com.wistrand.midlet.valentine;

import javax.microedition.lcdui.*;
import javax.microedition.midlet.*;

import java.util.*;

import com.wistrand.midlet.sprite.*;
import com.wistrand.midlet.fp.*;


/**
 * Sprite which looks like a heart and moves in a nice way.
 *
 * The heart will also create new BallonSprites and MultiStars
 * at random intervals.
 */
public class HeartSprite extends Sprite {

  SpriteCanvas canvas;

  int fw;
  int fh;

  int brightCol;
  int darkCol;

  int sizeD = 1;

  int minSize = FP.SCALE / 2;
  int maxSize = FP.SCALE * 2;

  Wobbler wobbler = new Wobbler(FP.SCALE, 240);
  Bouncer bouncer;

  public HeartSprite(SpriteCanvas canvas, int x, int y) {
    super();
    this.canvas = canvas;
    
    setSpeed(FP.rnd(- 3 * FP.SCALE, 3 * FP.SCALE),  FP.rnd(-FP.SCALE, FP.SCALE));
    
    moveTo(FP.SCALE * x, FP.SCALE * y);
    bouncer = new Bouncer(this.y, dy, 0, canvas.getHeight() << FP.SCALE, FP.SCALE * 2);

    size = FP.rnd(minSize, maxSize);

    setColor(0xff0000);

    fw = (canvas.getWidth() << FP.SCALE_BITS)  / 8;
    fh = (canvas.getHeight() << FP.SCALE_BITS) / 8;
  }

  public void moveTo(int x, int y) {
    super.moveTo(x, y);
  }

  public void setColor(int col) {
    super.setColor(col);

    brightCol = Util.interpolateRGB(col, 0xffffff, 128);
    darkCol   = Util.interpolateRGB(0x000000, col, 128);
  }

  public void update() {
    int w = canvas.getWidth()  << FP.SCALE_BITS;
    int h = (canvas.getHeight() << FP.SCALE_BITS) - fh * 2;

    wobbler.update();
    bouncer.max = h;
    bouncer.update();

    size += sizeD;
    
    if(size > maxSize) { sizeD = -sizeD; size = maxSize;}
    if(size < minSize) { sizeD = -sizeD; size = minSize;}
    
    super.update();
    
    
    if(x < 0) { dx = -dx; x = 0; }
    if(x > w) { dx = -dx; x = w; }

    y = bouncer.getValue();

    int bounce = bouncer.getBounce();
    if(bounce != 0) {
      wobbler.start(Math.max(-wobbler.max / 2, bounce / 16));
    }
    if(bouncer.meanY == h) {
      setSpeed(FP.rnd(-2 * FP.SCALE, 2 * FP.SCALE),  FP.rnd(-20 * FP.SCALE, 0));
      bouncer.delta = this.dy;
    }
    
    // Create ballon and stars at random intervals
    if((FP.rnd.nextInt() & 255) > 250) {
      BalloonSprite s = new BalloonSprite(canvas, Util.makeText(), x, y);
      canvas.addAfter(s, this);
    }
    if((FP.rnd.nextInt() & 255) > 253) {
      MultiStar s = new MultiStar(canvas.bgColor, canvas);
      s.setParent(this, 0, 0);
      canvas.addBefore(s, this);
    }
  }


  // Vertical lines describing the heart since MIDP doesn't 
  // have any fill primitive.
  int[] lines = {
    7, 13, // column 1
    5, 15, // column 2
    4, 17, // 3
    3, 18, // 4
    2, 19, // 5
    1, 20, // 6
    1, 21, // 7
    1, 22, // 8
    1, 23, // 9
    1, 24, // 10
    1, 24, // 11
    1, 25, // 12
    2, 26, // 13
    2, 26, // 14
    3, 27, // 15
    3, 27, // 15
    
  };
  
  /**
   * Paint a heart at <tt>(sx, sy)</tt> with height <tt>h</tt>
   */
  public void paintHeart(Graphics g, int sx, int sy, int h) {
    int w = fw >> FP.SCALE_BITS;
    
    w = w * size;
    w = w >> FP.SCALE_BITS;
    
    h = h * size;
    h = h >> FP.SCALE_BITS;

    for(int i = 0; i < w; i++) {
      int k  = lines.length / 2 * (w-i-1) / w;
      int y0 = h * lines[k * 2] / 32;
      int y1 = h * lines[k * 2 + 1] / 32;
      
      // Draw main body
      g.setColor(col);
      g.drawLine(sx - i, sy + y0, sx - i, sy + y1);
      g.drawLine(sx + i, sy + y0, sx + i, sy + y1);

      // Draw highlights
      if(i > w / 2) {
	g.setColor(brightCol);
	g.drawLine(sx - i, sy + y0, sx - i, sy + y0 + 1);
      } else {
	g.setColor(brightCol);
	g.drawLine(sx + i, sy + y0, sx + i, sy + y0 + 1);
      }
      
      // Draw shadow
      g.setColor(darkCol);
      g.drawLine(sx + i, sy + y1, sx + i, sy + y1 - 1);

    }
  }

  public void paint(Graphics g) {

    int sx = x >> FP.SCALE_BITS;
    int sy = y >> FP.SCALE_BITS;


    int fhD = (wobbler.getValue() * fh / 2) >> FP.SCALE_BITS;

    int w = fw >> FP.SCALE_BITS;
    int h = (fh + fhD) >> FP.SCALE_BITS;

    g.setColor(0xff0000);
    paintHeart(g, sx, sy, h * 2);
  }
}

